# Facer
# John W. Miller
# 2024

import sys

assert sys.version_info[0] == 3, "facer requires Python 3."

"""Simple Python face averaging"""

__author__ = "John W. Miller"
__url__ = "https://github.com/johnwmillr/facer"
__description__ = "Face averaging in Python made simple"
__license__ = "MIT"

print("Loading face detector and landmark prediction models...")
